/******************************************************************************
 * libreadpmud.c - library that reads from pmud socket, parses the output     *
 *                 and returns it                                             *
 *                                                                            *
 * Author: Steven Hanley                                                      *
 * Date: 13/01/2001                                                           *
 * See CVS log for changes                                                    *
 ******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <errno.h>
#include <pcre.h>
#include <stdio.h>
#include <unistd.h>

#include "libreadpmud.h"

#define SA struct sockaddr

static char *mfgets(char *str, int size, FILE *stream) {

   char *ch;
   int len;

   ch = fgets (str, size, stream);
   len = strlen (str);
   if (str[len - 1] == '\n') { 
      str[len - 1] = '\0';   
      len--;
      if (str[len - 1] == '\r') str[len - 1] = '\0';
   }

   return ch;
}

/* connects to the pmud socket on localhost and reads the first line
   retusn the line read */
char *readpm_read_from_pmud () {
   struct sockaddr_in serv;
   int sock;
   char line1[SLEN], line2[SLEN];
   char *ret;
   FILE *pmud;

   if ((sock = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
      fprintf (stderr, "error in socket call: %s\n", strerror(errno));
      exit (1);
   }

   memset (&serv, 0, sizeof(serv));
   serv.sin_family = AF_INET;
   serv.sin_port = htons (PMUD_PORT);

   if (inet_pton(AF_INET, PMUD_HOST, &serv.sin_addr) < 0) {
      fprintf (stderr, "inet_pton error for %s: %s\n", PMUD_HOST, strerror(errno));
      exit (1);
   }

   if (connect (sock, (SA *) &serv, sizeof(serv)) < 0) {
      fprintf (stderr, "error in connect call: %s\n", strerror(errno));
      exit (1);
   }

   pmud = fdopen (sock, "r");
   mfgets (line1, SLEN, pmud);
   mfgets (line2, SLEN, pmud);
   close (sock);
   ret = (char *) malloc ((strlen (line1) + strlen (line2) + 1) * sizeof (char));
   strncat (ret, line1, strlen (line1));
   strncat (ret, " ", 1);
   strncat (ret, line2, strlen (line2));
   return ret;
}

/* The pmud socket has this information printed ut when ever you connect to it
 * this is for the g3 only, as this is what I have
 * the version thing and the line of information are on two seperate lines, I combined 
 * them for this
 *
 *  pmud version somenumber? S {flags charge maxcharge ibat vbat} {(same for R batt)}
 *
 * Flags is a set of 3 bits, in binary, which we number left-to-right
 * from 0 to 2.  (Flag 0 is 1 in the byte we get from the PMU.)
 * Flag 0 indicates that the AC is connected.
 * Flag 1 indicates that the battery is being charged.
 * Flag 2 indicates that the battery is present.
 *
 * The ibat field gives the battery current, positive when charging.
 * The vbat field gives the battery voltage in millivolts maybe.
 *
 * example string (power plugged in and charging)
 * pmud 0.6 12 S {111 3868 4800 1348 12170} {110}
 * if there is a second batterry I guess all details will be there, however when there isnt
 * only the flags are there, so use the third flag to check if there is a second battery
 */

/* fills the bat from the list
   returns the number of fields from list it used */
static int fill_pminfo_battery (const char **list, readpm_battery *bat) {
   bat->ac = (list[0][0] == '1') ? 1 : 0;
   bat->charging = (list[1][0] == '1') ? 1 : 0;
   bat->battery = (list[2][0] == '1') ? 1 : 0;
   if (bat->battery) {
      /* we skip list[3] as this is the match for the line charge maxcharge current voltage
         when this line is present */
      bat->charge = strtol (list[4], NULL, 0);
      bat->maxcharge = strtol (list[5], NULL, 0);
      bat->current = strtol (list[6], NULL, 0);
      bat->voltage = strtol (list[7], NULL, 0);
      bat->percentage = bat->charge * 100 / bat->maxcharge;
      if (bat->current == 0) {
        bat->timeleft = 0; /* battery full and ac connected, time means nothing */
      } else {
         int timeleft = 0;
         if (bat->current < 0) { /* in use, negative */
            timeleft = (bat->charge * 3552) / (bat->current * -60);
         } else {
            timeleft = ((bat->maxcharge - bat->charge) * 3552) / (bat->current * 60);
         }
         bat->timeleft = timeleft;
      }
   }
   return 8; /* this isnt needed any more I think, it seems the behavior with pcre is
                in fact the same as perl, it returns empty array elements for empty 
                matched things */
}

/* returns an array of two pminfo structs, representing the first and second batteries */
readpm_pminfo *readpm_parse_pmstring (char *str) {
   pcre *cpat;
   int erroffset, execret, ovector[100], used;
   const char *pcreerr, **retlist;
   readpm_pminfo *pm;

   pm = (readpm_pminfo *) malloc (sizeof (readpm_pminfo));
   pm->string = str;
   pm->left.which = "left";
   pm->right.which = "right";

   cpat = pcre_compile (
             "pmud (\\d\\.?\\d?) (\\d+) S "           /* version and model */
             "\\{(\\d)(\\d)(\\d)"                     /* flags */
             "( (\\d+) (\\d+) (-?\\d+) (\\d+))?\\} "  /* charge maxcharge current voltage */
             "\\{(\\d)(\\d)(\\d)"                     /* flags for second baterry */
             "( (\\d+) (\\d+) (-?\\d+) (\\d+))?\\}",  /* as above */
             0, &pcreerr, &erroffset, NULL);
   /* with the above pattern the last 4 things in the matched subpatterns are empty if there is 
      no second battery */
   if ((execret = pcre_exec (cpat, NULL, str, strlen (str), 0, 0, ovector, 100)) < 0) {
      fprintf (stderr, "Execution of pattern failed: errno %d\n", execret);
      exit (1);
   }

   if ((pcre_get_substring_list (str, ovector, execret, &retlist))) {
      fprintf (stderr, "Failed to get substring list, probable cause: out of memory\n");
      exit (1);
   }

   if (execret == 0) {
      fprintf (stderr, "Error happened, execret is 0\n");
      return pm;
   }

   pm->version = strdup (retlist[1]);
   pm->model = strtol (retlist[2], NULL, 0);

   used = fill_pminfo_battery (&retlist[3], &(pm->left));
   fill_pminfo_battery (&retlist[3 + used], &(pm->right));

   pcre_free (cpat);
   pcre_free (retlist);

   return pm;
}

char *battery_info (readpm_battery *bat) {
   char line[SLEN], tmp[SLEN];
   *tmp = *line = '\0';

   if (! bat->battery) { return strdup (""); }

   snprintf (line, 9, "(%s (", bat->which);
   if (bat->charge == bat->maxcharge) {
      strncat (line, "full) ", 6);
   } else {
      if (bat->charging) {
         strncat (line, "charging) ", 10);
      } else {
         strncat (line, "in use) ", 9);
      }
      snprintf (tmp, 20, "%d%%, %02d:%02d ", bat->percentage, bat->timeleft/60, bat->timeleft%60);
      strncat (line, tmp, 20);
      snprintf (tmp, 10, " %dmA ", bat->current);
      strncat (line, tmp, 10);
   }

   snprintf (tmp, 8, "%.2fmV", (double) bat->voltage/1000);
   strncat (line, tmp, 8);
   strncat (line, ")", 1);

   /* the above size values add up to about 73 so they wont overflow the 100 char string 
      no matter what */
   return strdup(line);
}

char *readpm_parsed_output (readpm_pminfo *pm) {
  static char output[SLEN*2];
     char *left = battery_info (&(pm->left)), 
          *right = battery_info (&(pm->right));

      /* assume ac is plugged in if neither battery is present below, I think this is 
         a valid assumption <g> */
      snprintf (output, SLEN*2, "pmud %s, pb model %d, batteries %s%s%s", 
                pm->version, pm->model, 
                left,
                (pm->left.battery && pm->right.battery) ? " " : 
                (! (pm->left.battery || pm->right.battery)) ? "(none, ac plugged in)" : "",
                right);
      free (left); free (right);
      return output;
}

void readpm_free_pminfo (readpm_pminfo *pm) {
   free (pm->string);
   free (pm->version);
   free (pm);
}
